from typing import Optional

from irctk.client import Client
from irctk.command import Command
from irctk.message import Message, MessageTag


class Context:
    def __init__(
        self, sender: str, channel: Optional[str], message: Message, client: Client
    ):
        self.sender = sender
        self.channel = channel
        self.message = message
        self.client: Client = client

    def reply(self, text: str) -> None:
        if self.channel:
            target = self.channel
            text = f'{self.sender}: {text}'
        else:
            target = self.sender

        for line in text.splitlines():
            message = Message(command=Command.NOTICE.value, parameters=[target, line])
            msgid = self.message.find_tag('msgid')
            if msgid:
                message.tags.append(MessageTag(True, 'draft', 'reply', msgid))

            self.client.send(message)



import argparse
import asyncio
import importlib
import logging
import sys
import time
import tomllib

from irctk_bot.controller import Controller


def create_argument_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        'irctk-bot', description='irctk bot', add_help=False
    )

    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('-h', '--host')
    group.add_argument('-c', '--config', type=argparse.FileType('rb'))

    parser.add_argument('-m', '--module', action='append')
    parser.add_argument('-n', '--nick')
    parser.add_argument('-j', '--join', action='append')

    parser.add_argument('--verbose', action='store_true')
    return parser


def create_controller(args):
    host = None
    nick = 'bot'
    modules = set()
    channels = set()

    if args.config:
        config = tomllib.load(args.config)
        keys = set(config.keys())
        allowed_keys = set(['modules', 'host', 'nick', 'join'])
        for key in allowed_keys:
            if key in keys:
                keys.remove(key)
        if keys:
            print(f'Invalid fields in config: {keys}', file=sys.stderr)
            sys.exit(1)

        host = config['host']

        if 'nick' in config:
            nick = config['nick']

        if 'modules' in config:
            for module in config['modules']:
                modules.add(module)

        if 'join' in config:
            for channel in config['join']:
                channels.add(channel)

    if args.host:
        host = args.host

    if args.nick:
        nick = args.nick

    if args.join:
        for channel in args.join:
            channels.add(channel)

    if args.module:
        for module in args.module:
            modules.add(module)

    assert host
    controller = Controller(nick, host, channels)
    for module in modules:
        controller.add_module(importlib.import_module(module))

    return controller


def main() -> None:
    args = create_argument_parser().parse_args()

    if args.verbose:
        logging.basicConfig(level='DEBUG')
    else:
        logging.basicConfig(level='INFO')

    controller = create_controller(args)

    with asyncio.Runner() as runner:
        while True:
            try:
                runner.run(controller.connect())
            except Exception as e:
                print(e)

            # while not ideal, there does not appear to be public API to
            # allow us to differentiate between interrupt (abort process)
            # and the task exiting to trigger a reconnect.
            if getattr(runner, '_interrupt_count', 0) > 0:
                return

            time.sleep(2)



from inspect import Parameter, signature
from typing import Optional

import irctk
from irctk_bot.command import Command, TargetMode
from irctk_bot.context import Context
from irctk_bot.controller import Controller
from irctk_bot.module import Module

module = Module()


def command_parameters(func) -> Optional[str]:
    arguments = []

    for parameter in signature(func).parameters.values():
        if parameter.annotation in (Context, Controller, irctk.Client, irctk.Message):
            # irctk-bot injects these, not for user to input
            continue

        name = parameter.name.replace('_', ' ')

        if parameter.default == Parameter.empty:
            name = f'<{name}>'
        else:
            name = f'[{name}]'

        if parameter.kind == Parameter.VAR_POSITIONAL:
            name = f'{name}...'

        arguments.append(name)

    if arguments:
        return ' '.join(arguments)
    return None


def command_usage(command: Command) -> str:
    parameters = command_parameters(command.func)
    if parameters:
        return f'{command.name} {parameters}'
    return command.name


def show_command_in_help(context: Context):
    def evaluate(command: Command) -> bool:
        if command.target_mode == TargetMode.CHANNEL and not context.channel:
            return False

        if command.target_mode == TargetMode.NICK and context.channel:
            return False

        return command.show_help

    return evaluate


@module.command()
def help(
    context: Context, controller: Controller, command_name: Optional[str] = None
) -> None:
    if command_name:
        command = controller.find_command(command_name)
        if not command:
            context.reply(f'no such command: {command_name}')
            return

        if not command.func.__doc__:
            context.reply(f'no help available for the {command.name} command')
            return

        context.reply(command.func.__doc__.strip())

        parameters = command_parameters(command.func)
        if parameters:
            context.reply(f'Usage: {command.name} {parameters}')

        return

    commands = map(command_usage, filter(show_command_in_help(context), controller.commands))
    command = ', '.join(commands)
    context.reply(f'commands: {command}')



import asyncio
import logging
from typing import Any, Generator, Optional

import irctk
from irctk.client import Client
from irctk_bot.command import TargetMode
from irctk_bot.context import Context

logger = logging.getLogger(__name__)


class Controller:
    def __init__(self, nick: str, host: str, channels):
        self.clients = []
        self.modules = []

        self.nick = nick
        self.host = host
        self.channels = channels
        self.loop = asyncio.new_event_loop()

    async def connect(self) -> None:
        client = Client(self.nick)
        client.delegate = self
        self.clients.append(client)

        try:
            await client.connect(self.host, port=6697, use_tls=True)
        except asyncio.CancelledError:
            client.quit()
        finally:
            self.clients.remove(client)

    def add_module(self, module) -> None:
        self.modules.append(module)

    def find_module(self, name: str):
        for module in self.modules:
            if module.__name__ == name:
                return module

        return None

    @property
    def commands(self) -> Generator[Any, None, None]:
        for module in self.modules:
            for command in module.module.commands:
                yield command

    def irc_command_handlers(self, command: str):
        for module in self.modules:
            if command not in module.module.irc_commands:
                continue
            for handler in module.module.irc_commands[command]:
                yield handler

    def find_command(self, name: str) -> Optional[Any]:
        for command in self.commands:
            if command.name == name:
                return command

        return None

    def run_command_handlers(
        self, client: irctk.Client, message: irctk.Message
    ) -> None:
        for handler in self.irc_command_handlers(message.command):
            try:
                handler(client=client, message=message)
            except Exception as e:
                logger.error('Cannot invoke %s', handler, e)

    # ClientDelegate

    def irc_disconnected(self, client, error):
        logger.error('Disconnected %s', error)

    def irc_registered(self, client):
        client.send('MODE', client.nick, '+B')
        for channel in self.channels:
            client.send_join(channel)

    def irc_channel_quit(
        self, client: irctk.Client, nick, channel, reason: Optional[str]
    ) -> None:
        if nick.nick == self.nick and client.nick.nick != self.nick:
            client.send('NICK', self.nick)

    def irc_message(self, client: irctk.Client, message: irctk.Message) -> None:
        self.run_command_handlers(client, message)

        if message.command != 'PRIVMSG':
            return

        if not message.prefix:
            return

        sender = client.nick_class.parse(message.prefix).nick
        target = message.get(0)
        text = message.get(1)
        if not target or not text:
            return

        if client.irc_equal(target, client.nick.nick):
            context = Context(sender, None, message, client)
        else:
            context = Context(sender, target, message, client)

        if context.channel and not text.startswith(f'{client.nick.nick}: '):
            return

        if text.startswith(f'{client.nick.nick}: '):
            text = text[len(client.nick.nick) + 2 :]

        command_name, _, args = text.partition(' ')
        command = self.find_command(command_name)
        if not command:
            context.reply(f'command {command_name} not found')
            return

        if command.target_mode == TargetMode.CHANNEL:
            if not context.channel:
                return context.reply(f'{command_name} can only be used in a channel')
        elif command.target_mode == TargetMode.NICK:
            if context.channel:
                return context.reply(f'{command_name} can only be used in DM')
        elif command.target_mode != TargetMode.ALL:
            logger.error('Unexpected target mode %s', command.target_mode)
            context.reply('internal error')
            return

        try:
            args, kwargs = command.parser.parse(
                args,
                dependencies={
                    Context: context,
                    irctk.Message: message,
                    Client: client,
                    Controller: self,
                },
            )
        except ValueError as e:
            if len(e.args) == 2:
                context.reply(f'{e.args[0]}: {e.args[1]}')
            else:
                context.reply(str(e))
            return
        except TypeError as e:
            logger.error('Cannot invoke %s', command.func, e)
            context.reply('internal error')
            return

        logger.info('Running %s (%s %s)', command.func, args, kwargs)
        try:
            command.func(*args, **kwargs)
        except Exception as e:
            logger.error('Cannot invoke %s', command.func, e)
            context.reply('internal error')



from collections import defaultdict
from typing import Any, Dict, List, Optional

from irctk_bot.argument_parser import ArgumentParser
from irctk_bot.command import Command, TargetMode


class Module:
    def __init__(self):
        self.commands = []
        self.irc_commands: Dict[str, List[Any]] = defaultdict(list)

    def command(
        self,
        name: Optional[str] = None,
        show_help: bool = True,
        target_mode: TargetMode = TargetMode.ALL,
    ):
        """
        Register a command with the module.

        :param name: The name of the command, this will default to the function name if unset.
        :param show_help: Whether to show the command in help list output.
        :param target_mode: The targets that the command can be used within.

        >>> @module.command()
        >>> def ping(context: Context):
        >>>    context.reply('pong')
        """

        def register(func):
            parser = ArgumentParser(func)
            parser.validate()

            self.commands.append(
                Command(
                    name or func.__name__,
                    func,
                    parser,
                    show_help=show_help,
                    target_mode=target_mode,
                )
            )
            return func

        return register

    def irc_command(self, command: str):
        """
        Register a hook for any IRC message.

        >>> @module.irc_command('PRIVMSG')
        >>> def echo(client: irctk.Client, message: irctk.Message):
        >>>    client.send_privmsg(message.prefix.nick, message.get(2))
        """

        def register(func):
            self.irc_commands[command].append(func)
            return func

        return register



from inspect import Parameter, signature
from typing import Any, Dict, List, Optional, Tuple, Union

from irctk.client import Client
from irctk.message import Message
from irctk_bot.context import Context
from irctk_bot.controller import Controller


class ArgumentBuilder:
    def __init__(self):
        self.args = []
        self.kwargs = {}

    def add(self, value: Any, parameter: Parameter) -> None:
        if parameter.kind == Parameter.KEYWORD_ONLY:
            self.kwargs[parameter.name] = value
        elif parameter.kind == Parameter.POSITIONAL_OR_KEYWORD:
            self.args.append(value)

    def build(self) -> Tuple[List[str], Dict[str, Any]]:
        return (self.args, self.kwargs)


class ArgumentParser:
    def __init__(self, func):
        self.func = func
        self.supported_types = (
            str,
            int,
            Context,
            Any,
            Parameter.empty,
            Controller,
            Client,
            Message,
        )

    def validate(self):
        for value in signature(self.func).parameters.values():
            if value.kind == Parameter.POSITIONAL_OR_KEYWORD:
                if (
                    self.unwrap_type_optional(value.annotation)
                    not in self.supported_types
                ):
                    raise TypeError(f'unsupported type {value.annotation}')

            elif value.kind == Parameter.VAR_POSITIONAL:
                pass

            else:
                raise TypeError(f'unsupported argument kind {value.kind}')

    def unwrap_type_optional(self, annotation):
        if '__origin__' not in annotation.__dict__:
            return annotation

        if '__args__' not in annotation.__dict__:
            return annotation

        if annotation.__dict__['__origin__'] != Union:
            return annotation

        args = list(annotation.__dict__['__args__'])
        args.remove(None.__class__)
        if len(args) == 1:
            return args[0]
        return annotation

    def parse(
        self, text: str, dependencies: Optional[Dict[Any, Any]] = None
    ) -> Tuple[List[str], Dict[str, Any]]:
        builder = ArgumentBuilder()

        for parameter in signature(self.func).parameters.values():
            human_name = parameter.name.replace('_', ' ')

            if parameter.kind == Parameter.POSITIONAL_OR_KEYWORD:
                if dependencies and parameter.annotation in dependencies:
                    builder.add(dependencies[parameter.annotation], parameter)

                elif len(text):
                    value, _, text = text.partition(' ')

                    annotation = self.unwrap_type_optional(parameter.annotation)

                    if parameter.annotation == Parameter.empty:
                        builder.add(value, parameter)
                    elif parameter.annotation == Any:
                        builder.add(value, parameter)
                    elif annotation == str:
                        builder.add(value, parameter)
                    elif annotation == int:
                        try:
                            integer_value = int(value)
                        except ValueError:
                            raise ValueError(
                                human_name, 'argument is not a valid number'
                            )
                        builder.add(integer_value, parameter)
                    else:
                        raise TypeError(f'unsupported type {parameter.annotation}')
                elif parameter.default == Parameter.empty and not len(text):
                    raise ValueError(human_name, 'missing argument')

            elif parameter.kind == Parameter.VAR_POSITIONAL:
                builder.args += text.split()
                text = ''

            else:
                raise TypeError(f'unsupported argument kind {parameter.kind}')

        if len(text):
            raise ValueError(f'unexpected arguments {text}')

        return builder.build()



from enum import Enum, auto
from functools import wraps
from typing import Optional


class TargetMode(Enum):
    ALL = auto()
    CHANNEL = auto()
    NICK = auto()


class Command:
    def __init__(
        self,
        name: str,
        func,
        parser,
        show_help: bool = True,
        target_mode: TargetMode = TargetMode.ALL,
    ):
        self.name = name
        self.func = func
        self.parser = parser
        self.show_help = show_help
        self.target_mode = target_mode


def requires_account(account: Optional[str] = None):
    def wrap(func):
        @wraps(func)
        def inner(context, *args, **kwargs):
            if account:
                if context.message.account != account:
                    return context.reply('unauthorized')

            elif not context.message.account:
                return context.reply('unauthorized, please login')

            return func(context, *args, **kwargs)

        return inner

    return wrap
